from .injector import component as component
from .injector import inject as inject
from .injector import get as get
from .injector import single as single
from .injector import factory as factory
from .injector import auto_single as auto_single
from .injector import auto_factory as auto_factory

__version__ = "0.0.4"
